import argparse
import os
import shutil
import yaml
from os.path import dirname, join, exists

import starmatrix
import starmatrix.settings as settings
import starmatrix.model as model


def main():
    parser = argparse.ArgumentParser(
        prog="starmatrix",
        description="Command line tools for the Starmatrix Python library.",
        epilog="Another dimension, new Galaxy!",
    )
    parser.add_argument("-v", "--version", action="version", version=starmatrix.__version__)
    parser.add_argument("--config", metavar="FILENAME", help="configuration file to use containing model initial params")
    parser.add_argument("--generate-config", action="store_true", help="create a config.yml example file")

    args = parser.parse_args()

    if args.generate_config:
        return create_template_config_file()

    input_params = {}
    if args.config is not None:
        input_params = read_config_file(args.config)

    context = settings.validate(input_params)

    print("Running model with settings:")
    print("")
    for param in context:
        print("   " + str(param) + " = " + str(context[param]))
    print("")

    create_output_directory(context['output_dir'])

    model.Model(context).run()
    print(f"Done. Output files ready in '{context['output_dir']}' directory.")


def create_output_directory(output_dir):
    shutil.rmtree(output_dir, ignore_errors=True)
    if not exists(output_dir):
        os.makedirs(output_dir)


def create_template_config_file():
    shutil.copy(join(dirname(__file__), "sample_input", "params.yml"), "config-example.yml")
    return "Created file: config-example.yml"


def read_config_file(name):
    with open(name, "r") as params_file:
        input_params = yaml.safe_load(params_file)
    return input_params



"""
Chemical Abundances

Contains solar abundances from different papers/authors:

* Anders & Grevesse 1989,
* Grevesse & Sauval 1998,
* Asplund et al. 2005,
* Asplund et al. 2009,
* Heger 2010
* Lodders et al. 2019

and a way to define new abundances subclassing Abundances

"""

import math


def select_abundances(option, z):
    abundandes_data = {
        "ag89": AndersGrevesse1989,
        "gs98": GrevesseSauval1998,
        "as05": Asplund2005,
        "as09": Asplund2009,
        "he10": Heger2010,
        "lo19": Lodders2019
    }
    return abundandes_data[option](z)


class Abundances:
    def __init__(self, z):
        self.z = z

    def description(self):
        return "Base Abundances class"

    def feh(self):
        if self.z == 0:
            return -20
        else:
            return self.feh_z_non_zero()

    def abundance(self):
        return {
            "H":   self.h(),
            "D":   self.elements()["D"],
            "He3": self.elements()["He3"],
            "He4": self.he4(),
            "C":   self.elements()["C"] * (10 ** self.feh()),
            "C13": self.elements()["C13"] * (10 ** self.feh()),
            "N":   self.elements()["N"] * (10 ** self.feh()),
            "O":   self.elements()["O"] * (10 ** self.feh()),
            "Ne":  self.elements()["Ne"] * (10 ** self.feh()),
            "Mg":  self.elements()["Mg"] * (10 ** self.feh()),
            "Si":  self.elements()["Si"] * (10 ** self.feh()),
            "S":   self.elements()["S"] * (10 ** self.feh()),
            "Ca":  self.elements()["Ca"] * (10 ** self.feh()),
            "Fe":  self.elements()["Fe"] * (10 ** self.feh())
        }

    def corrected_abundance_CRI_LIM(self):
        """
        When using the combination of yields from
        Cristallo et al. 2011 (for low mass stars) + Limongi & Chieffi 2012 (for massive stars)
        data needs to be corrected because for non solar metalicities they don't follow solar scale.
        """
        if self.z < 0.014:
            return {
                "H":   self.h(),
                "D":   self.elements()["D"],
                "He3": self.elements()["He3"],
                "He4": self.he4(),
                "C":   self.elements()["C"] * (10 ** self.feh()),
                "C13": self.elements()["C13"] * (10 ** self.feh()),
                "N":   self.elements()["N"] * (10 ** self.feh()),
                "O":   self.elements()["O"] * (10 ** (self.feh() + 0.47)),
                "Ne":  self.elements()["Ne"] * (10 ** self.feh()),
                "Mg":  self.elements()["Mg"] * (10 ** (self.feh() + 0.27)),
                "Si":  self.elements()["Si"] * (10 ** (self.feh() + 0.37)),
                "S":   self.elements()["S"] * (10 ** (self.feh() + 0.35)),
                "Ca":  self.elements()["Ca"] * (10 ** (self.feh() + 0.33)),
                "Fe":  self.elements()["Fe"] * (10 ** self.feh())
            }
        else:
            return self.abundance()


class AndersGrevesse1989(Abundances):
    def h(self):
        return -2.398 * self.z + 0.7516

    def he4(self):
        return 1.40 * self.z + 0.2484

    def feh_z_non_zero(self):
        return math.log10(self.z / 0.02)

    def elements(self):
        return {
            "H":   0.705e-0,
            "D":   4.840e-5,
            "He3": 2.900e-5,
            "He4": 2.750e-1,
            "C":   3.000e-3,
            "C13": 2.830e-5,
            "N":   1.100e-3,
            "O":   8.600e-3,
            "Ne":  1.690e-3,
            "Mg":  5.200e-4,
            "Si":  6.500e-4,
            "S":   4.000e-4,
            "Ca":  6.000e-5,
            "Fe":  1.200e-3
        }

    def description(self):
        return "Anders & Grevesse 1989"


class GrevesseSauval1998(Abundances):
    def h(self):
        return -2.33 * self.z + 0.7516

    def he4(self):
        return 1.33 * self.z + 0.2484

    def feh_z_non_zero(self):
        return math.log10(self.z / 0.02)

    def elements(self):
        return {
            "H":   0.733e-0,
            "D":   4.800e-5,
            "He3": 2.900e-5,
            "He4": 2.495e-1,
            "C":   2.890e-3,
            "C13": 2.830e-5,
            "N":   8.480e-4,
            "O":   7.880e-3,
            "Ne":  1.750e-3,
            "Mg":  6.650e-4,
            "Si":  7.240e-4,
            "S":   4.980e-4,
            "Ca":  6.680e-5,
            "Fe":  1.290e-3
        }

    def description(self):
        return "Grevesse & Sauval 1998"


class Asplund2005(Abundances):
    def h(self):
        return -1.092 * self.z + 0.75157

    def he4(self):
        return 0.06 * self.z + 0.2484

    def feh_z_non_zero(self):
        return math.log10(self.z / 0.012)

    def elements(self):
        return {
            "H":   0.739e-0,
            "D":   4.800e-5,
            "He3": 2.900e-5,
            "He4": 2.486e-1,
            "C":   2.180e-3,
            "C13": 2.830e-5,
            "N":   6.240e-4,
            "O":   5.400e-3,
            "Ne":  1.020e-3,
            "Mg":  6.010e-4,
            "Si":  6.700e-4,
            "S":   3.270e-4,
            "Ca":  6.040e-5,
            "Fe":  1.170e-3
        }

    def description(self):
        return "Asplund et al. 2005"


class Asplund2009(Abundances):
    def h(self):
        return -1.06167 * self.z + 0.7524

    def he4(self):
        return 0.067 * self.z + 0.2476

    def feh_z_non_zero(self):
        return math.log10(self.z / 0.0134)

    def elements(self):
        return {
            "H":   0.738e00,
            "D":   1.430e-5,
            "He3": 4.490e-5,
            "He4": 2.485e-1,
            "C":   2.384e-3,
            "C13": 2.830e-5,
            "N":   6.986e-4,
            "O":   5.784e-3,
            "Ne":  1.256e-3,
            "Mg":  7.052e-4,
            "Si":  6.688e-4,
            "S":   3.114e-4,
            "Ca":  6.459e-5,
            "Fe":  1.307e-3
        }

    def description(self):
        return "Asplund et al. 2009"


class Heger2010(Abundances):
    def h(self):
        return -2.6866 * self.z + 0.7513

    def he4(self):
        return 1.687 * self.z + 0.2487

    def feh_z_non_zero(self):
        return math.log10(self.z / 0.015)

    def elements(self):
        return {
            "H":   0.711e0,
            "D":   2.76e-5,
            "He3": 3.41e-5,
            "He4": 2.74e-1,
            "C":   2.46e-3,
            "C13": 2.98e-5,
            "N":   7.96e-4,
            "O":   6.60e-3,
            "Ne":  1.17e-3,
            "Mg":  5.65e-4,
            "Si":  7.55e-4,
            "S":   3.96e-4,
            "Ca":  7.13e-5,
            "Fe":  1.26e-3
        }

    def description(self):
        return "Heger 2010"


class Lodders2019(Abundances):
    def h(self):
        return -2.6866 * self.z + 0.7513

    def he4(self):
        return 1.687 * self.z + 0.2487

    def feh_z_non_zero(self):
        return math.log10(self.z / 0.015)

    def elements(self):
        return {
            "H":   0.7048,
            "D":   2.78e-5,
            "He3": 3.47e-5,
            "He4": 0.2786,
            "C":   3.031e-3,
            "C13": 3.432e-5,
            "N":   8.536e-4,
            "O":   7.428e-3,
            "Ne":  2.271e-3,
            "Mg":  5.437e-4,
            "Si":  7.220e-4,
            "S":   3.719e-4,
            "Ca":  6.209e-5,
            "Fe":  1.255e-3
        }

    def description(self):
        return "Lodders et al. 2019"



import pytest
import math
import numpy as np
import starmatrix.settings as settings
from starmatrix.imfs import select_imf, IMF
from starmatrix.imfs import Salpeter, Starburst, Chabrier, Ferrini, Kroupa, MillerScalo, Maschberger


@pytest.fixture
def available_imfs():
    """
    Fixture returning the names of all available IMFs defined in settings
    """
    return settings.valid_values["imf"]


def test_select_imf():
    strings = ["salpeter", "starburst", "chabrier", "ferrini", "kroupa", "miller_scalo", "maschberger"]
    classes = [Salpeter, Starburst, Chabrier, Ferrini, Kroupa, MillerScalo, Maschberger]

    for i in range(len(strings)):
        imf_instance = select_imf(strings[i])
        assert type(imf_instance) == classes[i]


def test_valid_values_presence(available_imfs):
    for imf in available_imfs:
        assert select_imf(imf) is not None


def test_description_presence(available_imfs):
    for imf in available_imfs:
        assert select_imf(imf).description() != IMF().description()


def test_salpeter_alpha():
    salpeter = Salpeter({"imf_alpha": 2.33})
    assert salpeter.alpha() == 2.33


def test_imf_is_zero_if_no_positive_mass(available_imfs):
    for imf in available_imfs:
        assert select_imf(imf).for_mass(0) == 0.0
        assert select_imf(imf).for_mass(-10) == 0.0

        for mass in [0.015, 0.02, 0.2, 0.75, 2, 8, 35, 90]:
            assert select_imf(imf).for_mass(mass) > 0.0


def test_minimum_mass_value_for_kroupa_imf():
    assert select_imf("kroupa").for_mass(0.014) == 0.0
    assert select_imf("kroupa").for_mass(0.015) > 0.0


def test_for_mass_is_normalized(available_imfs):
    for imf_name in available_imfs:
        imf = select_imf(imf_name)
        mass = np.random.random() * 10
        imf_for_mass = imf.for_mass(mass)
        assert imf_for_mass == imf.m_phi(mass) * imf.normalization_factor


def test_normalization_factor():
    imf_1 = IMF({"imf_m_low": 7, "imf_m_up": 47})
    imf_2 = IMF({"imf_m_low": 7, "imf_m_up": 57})
    assert imf_2.integrated_m_phi_in_mass_interval() > imf_1.integrated_m_phi_in_mass_interval()
    assert imf_1.normalization_factor > imf_2.normalization_factor
    assert imf_1.integrated_m_phi_in_mass_interval() == 1 / imf_1.normalization_factor
    assert imf_2.integrated_m_phi_in_mass_interval() == 1 / imf_2.normalization_factor


def test_phi_m_phi_relation(available_imfs):
    for imf in available_imfs:
        selected_imf = select_imf(imf)
        for mass in [0.015, 0.02, 0.2, 0.75, 2, 8, 35, 90, np.random.random() * 10]:
            assert selected_imf.phi(mass) == selected_imf.m_phi(mass) / mass


def test_stars_per_mass_unit(available_imfs):
    for imf in available_imfs:
        selected_imf = select_imf(imf)
        expected = selected_imf.normalization_factor * selected_imf.integrated_phi_in_mass_interval()
        assert selected_imf.stars_per_mass_unit == expected



"""
Delay Time Distributions

Contains some predefined DTDs from different papers/authors:

* Ruiz Lapuente & Canal (2000)
* Maoz & Graur (2017)
* Castrillo et al (2020)
* Greggio, L. (2005)
* Strolger et al. (2020)

"""

import math
import scipy.integrate
import starmatrix.constants as constants
from functools import lru_cache


def select_dtd(option):
    dtds = {
        "rlp": dtd_ruiz_lapuente,
        "maoz": dtd_maoz_graur,
        "castrillo": dtd_castrillo,
        "greggio": dtd_greggio,
        "greggio-CDD04": dtd_close_dd_04,
        "greggio-CDD1": dtd_close_dd_1,
        "greggio-WDD04": dtd_wide_dd_04,
        "greggio-WDD1": dtd_wide_dd_1,
        "greggio-SDCH": dtd_sd_chandra,
        "greggio-SDSCH": dtd_sd_subchandra,
        "chen": dtd_chen,
        "strolger-fit1": dtds_strolger["fit_1"],
        "strolger-fit2": dtds_strolger["fit_2"],
        "strolger-fit3": dtds_strolger["fit_3"],
        "strolger-fit4": dtds_strolger["fit_4"],
        "strolger-optimized": dtds_strolger["optimized"],
    }
    return dtds[option]


def dtd_correction(params):
    """
    When normalizing DTDs to 1 sometimes a correction factor is needed,
    caused by the uncertainty in the value of the integral of the DTD for the whole mass range

    """
    if "dtd_correction_factor" in params:
        return params["dtd_correction_factor"]

    return 1.0


def dtd_ruiz_lapuente(t):
    """
    Delay Time Distribution (DTD) from Ruiz Lapuente & Canal (2000)

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9
    if logt < 7.8:
        return 0.0

    f1 = 0.170e-11 * math.exp(-0.5 * ((logt - 7.744) / 0.08198) ** 2)
    f2 = 0.338e-11 * math.exp(-0.5 * ((logt - 7.9867) / 0.12489) ** 2)
    f3 = 0.115e-11 * math.exp(-0.5 * ((logt - 8.3477) / 0.14675) ** 2)
    f4 = 0.160e-11 * math.exp(-0.5 * ((logt - 9.08) / 0.23) ** 2)
    f5 = 0.020e-11 * math.exp(-0.5 * ((logt - 9.58) / 0.17) ** 2)

    dtd = (f1 + f2 + f3 + f4 + f5) * 1e9

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    rate = 0.2440759  # [SN / Yr / M*]

    return rate * dtd


def dtd_maoz_graur(t):
    """
    Delay Time Distribution (DTD) from Maoz & Graur (2017)

    """
    if t <= 0.05:
        return 0.0

    dtd = math.pow(t, -1.1)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    rate = 1.793e-4  # [SN / Yr / M*]

    return rate * dtd


def dtd_castrillo(t):
    """
    Delay Time Distribution (DTD) from Castrillo et al (2020, in preparation)

    """
    if t <= 0.04:
        return 0.0

    dtd = math.pow(t, -1.2)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    rate = 1.5879e-4  # [SN / Yr / M*]

    return rate * dtd


def dtd_greggio(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9
    if logt < 7.45:
        dtd = 0
    elif 7.45 <= logt <= 7.735:
        dtd = (0.00215/(7.776-7.516)) * (logt-7.50)
    elif 7.735 < logt <= 8.55:
        dtd = 0.003335 * math.exp(((-0.5 * (logt-8.22))/0.47) ** 2)
    elif 8.55 < logt <= 8.61:
        dtd = 0.002618 - ((0.002618-0.001129)/(8.6398-8.55)) * (logt-8.55)
    elif 8.61 < logt:
        dtd = math.pow(10, ((-1.0615 * logt) + 6))

    # Normalization
    return dtd * 0.524563739


def dtd_close_dd_04(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)
    Model Close Double Degenerate 0.4 Gyrs

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9
    if logt < 7.657:
        log_dtd = -20
    elif 7.657 <= logt < 8.6:
        log_dtd = -0.8373*(logt**2) + 13.217*logt - 51.878
    elif 8.6 <= logt:
        log_dtd = 10.914 - 1.2964*logt

    dtd = math.pow(10, log_dtd)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    return dtd * 1.059e-3


def dtd_close_dd_1(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)
    Model Close Double Degenerate 1 Gyr

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9

    if logt < 6.32:
        log_dtd = -20
    elif 6.32 <= logt < 7.9:
        log_dtd = 4.68e-3 + 4.86e-2*(logt-6.32)
    elif 7.9 <= logt < 8.987:
        log_dtd = 4.117 - 0.5092*logt
    elif 8.987 <= logt < 9.16:
        log_dtd = -97.15408482273*(logt**3) + 2653.445666247*(logt**2) - 24157.97809549*logt + 73317.50108578
    elif 9.16 <= logt:
        log_dtd = 8.8761 - 1.0656*logt

    dtd = math.pow(10, log_dtd)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    return dtd * 9.8400244741154628e-4


def dtd_wide_dd_04(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)
    Model Wide Double Degenerate 0.4 Gyrs

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9
    if logt < 7.5:
        log_dtd = -20
    elif 7.5 <= logt < 8.746:
        log_dtd = -1.194384*(logt**4) + 37.542520*(logt**3) - 442.490023*(logt**2) + 2318.164371*logt - 4555.628418
    elif 8.746 <= logt:
        log_dtd = -0.8223*logt + 6.6439

    dtd = math.pow(10, log_dtd)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    return dtd * 9.92088e-4


def dtd_wide_dd_1(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)
    Model Wide Double Degenerate 1 Gyr

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9
    if logt < 7.69:
        log_dtd = -20
    elif 7.69 <= logt < 8.99:
        log_dtd = 0.1305*(logt**4) - 4.295277*(logt**3) + 52.687732*(logt**2) - 285.362875*logt + 575.188754
    elif 8.99 <= logt:
        log_dtd = 1.061260*(logt**4) - 41.103454*(logt**3) + 596.908266*(logt**2) - 3852.941647*logt + 9327.789697

    dtd = math.pow(10, log_dtd)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    return dtd * 1.1169615977359982e-3


def dtd_sd_chandra(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)
    Model Single Degenerate Chandra Mass

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9

    if logt < 7.89:
        log_dtd = -20
    elif 7.89 <= logt < 9.1:
        log_dtd = -0.0869*(logt**3) + 1.9168*(logt**2) - 14.187*logt + 35.319
    elif 9.1 <= logt < 9.89:
        log_dtd = -1.7291*logt + 15.144
    elif 9.89 <= logt:
        log_dtd = -5.1962*logt + 49.362

    dtd = math.pow(10, log_dtd)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    return dtd * 1.064741165931863e-3


def dtd_sd_subchandra(t):
    """
    Delay Time Distribution (DTD) from Laura Greggio, A&A 441, 1055–1078 (2005)
    Model Single Degenerate Sub-Chandra Mass

    """
    if t <= 0:
        return 0.0

    logt = math.log10(t) + 9
    if logt < 7.60:
        log_dtd = -20
    elif 7.60 <= logt < 8.58:
        log_dtd = 0.2564*(logt**3) - 6.8315*(logt**2) + 59.975*logt - 173.57
    elif 8.58 <= logt:
        log_dtd = -1.733*logt + 14.852

    dtd = math.pow(10, log_dtd)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    return dtd * 1.03540652488714e-3


def dtd_chen(t):
    """
    Delay Time Distribution (DTD) from Chen, Hu and Wang, 2021, ApJ

    """
    if t <= 0.12:
        return 0.0

    dtd = math.pow(t, -1.41)

    # Normalization using 1.03e-3 SN/M* as Hubble-time-integrated production efficiency SN/Mo
    rate = 2.069e-4  # [SN / Yr / M*]

    return rate * dtd


class Strolger:
    def __init__(self, psi, omega, alpha):
        self.psi = psi
        self.omega = omega
        self.alpha = alpha

    def description(self):
        return ("Delay Time Distributions (DTDs) from Strolger et al, "
                "The Astrophysical Journal, 2020, 890, 2. "
                "DOI: 10.3847/1538-4357/ab6a97")

    def phi(self, t_gyrs):
        t_myrs = t_gyrs * 1e3

        u = t_myrs - self.psi
        term_1 = (1/(self.omega * math.pi)) * math.exp((-(u**2))/(2*(self.omega**2)))

        t_low = -math.inf
        t_up = self.alpha*(u/self.omega)

        if 12 < t_up:
            term_2 = 2 * scipy.integrate.quad(self.term_2_f, t_low, 0)[0]
        else:
            term_2 = scipy.integrate.quad(self.term_2_f, t_low, t_up)[0]

        return term_1 * term_2

    def term_2_f(self, t_prime):
        return math.exp(-math.pow(t_prime, 2)/2)

    @lru_cache(maxsize=128)
    def normalization_rate(self):
        return self.efficiency() / self.phi_integrated()

    def efficiency(self):
        # SN/M* as Hubble-time-integrated production efficiency SN/Mo
        return 1.03e-3

    def phi_integrated(self):
        return scipy.integrate.quad(self.phi, 0, constants.TOTAL_TIME)[0]

    def at_time(self, t):
        return self.normalization_rate() * self.phi(t)


dtds_strolger = {
    "fit_1": Strolger(10, 600, 220).at_time,
    "fit_2": Strolger(110, 1000, 2).at_time,
    "fit_3": Strolger(350, 1200, 20).at_time,
    "fit_4": Strolger(6000, 6000, -2).at_time,
    "optimized": Strolger(-1518, 51, 50).at_time,
}



import pytest
from pytest_mock import mocker
import numpy as np
import starmatrix.constants as constants
import starmatrix.settings as settings
import starmatrix.elements as elements
import starmatrix.abundances as abundances
import starmatrix.matrix as matrix


def test_q_index():
    assert matrix.q_index("H") == 0
    assert matrix.q_index("Fe") == 14


def resize_matrix():
    expected_size = (constants.Q_MATRIX_ROWS, constants.Q_MATRIX_COLUMNS)
    assert np.zeros((constants.Q_MATRIX_ROWS + 5, constants.Q_MATRIX_COLUMNS + 33)).shape == expected_size


def test_empty_q_matrix():
    empty_matrix = matrix.empty_q_matrix()

    assert empty_matrix.shape == (15, 15)
    assert np.all([i == 0 for i in empty_matrix])


def test_matrices_are_empty_if_not_enough_mass():
    m = constants.M_MIN - 0.001
    test_settings = {
        "z": 0.03,
        "abundances": abundances.select_abundances(np.random.choice(settings.valid_values["sol_ab"]), 0.03),
        "expelled": elements.Expelled(settings.default["expelled_elements_filename"]),
    }

    assert np.all([i == 0 for i in matrix.q(m, test_settings)])
    assert np.all([i == 0 for i in matrix.q_sn(m)])


def test_q_sn_size():
    for m in [0.8, 1, 2, 4, 6, (np.random.rand() * 8), 8, 10, 40, 90]:
        for feh in [-3.3, -1.3, -0.3, 0., 0.17, 0.3, 0.4]:  # test z from 0.00001 to 0.05

            q_supernovae = matrix.q_sn(m, feh=feh)

            assert q_supernovae.shape == (constants.Q_MATRIX_ROWS, constants.Q_MATRIX_COLUMNS)


def test_q_size():
    for m in [0.8, 1, 2, 4, 6, 8, 10, 40, 90]:
        for z in [0., 0.001, 0.01, 0.02, 0.03, 0.04, 0.05]:
            test_settings = {
                "z": z,
                "abundances": abundances.select_abundances(np.random.choice(settings.valid_values["sol_ab"]), z),
                "expelled": elements.Expelled(settings.default["expelled_elements_filename"]),
            }

            q = matrix.q(m, test_settings)

            assert q.shape == (constants.Q_MATRIX_ROWS, constants.Q_MATRIX_COLUMNS)


def test_q_applies_yield_corrections(mocker):
    m = constants.M_MIN + np.random.rand() * 40
    expelled = elements.Expelled(settings.default["expelled_elements_filename"])
    test_settings = {
                "z": 0.02,
                "abundances": abundances.select_abundances(np.random.choice(settings.valid_values["sol_ab"]), 0.02),
                "expelled": expelled,
                "yield_corrections": {"Mg": 2, "Fe": 3.45}
            }

    mocker.spy(expelled, "for_mass")
    q = matrix.q(m, test_settings)
    expelled.for_mass.assert_called_once_with(m, {"Mg": 2, "Fe": 3.45})


def test_q_with_no_yield_corrections(mocker):
    m = constants.M_MIN + np.random.rand() * 40
    expelled = elements.Expelled(settings.default["expelled_elements_filename"])
    test_settings = {
                "z": 0.02,
                "abundances": abundances.select_abundances(np.random.choice(settings.valid_values["sol_ab"]), 0.02),
                "expelled": expelled,
            }

    mocker.spy(expelled, "for_mass")
    q = matrix.q(m, test_settings)
    expelled.for_mass.assert_called_once_with(m, {})


def test_cri_lim_exception(mocker):
    test_settings = {
        "z": 0.03,
        "abundances": abundances.select_abundances(np.random.choice(settings.valid_values["sol_ab"]), 0.03),
        "expelled": elements.Expelled(settings.default["expelled_elements_filename"]),
    }
    mocker.spy(abundances.Abundances, "abundance")
    mocker.spy(abundances.Abundances, "corrected_abundance_CRI_LIM")

    q = matrix.q(4, test_settings)

    abundances.Abundances.abundance.assert_called_once()
    abundances.Abundances.corrected_abundance_CRI_LIM.assert_not_called()

    test_settings["expelled"].cri_lim_yields = True

    q = matrix.q(4, test_settings)

    abundances.Abundances.corrected_abundance_CRI_LIM.assert_called_once()



"""
Default settings and initial parameters

All the values set here can be overwritten via the input file: params.yml

"""
from os.path import dirname, join
from starmatrix import constants as constants
from starmatrix import elements
from starmatrix.functions import max_mass_allowed

default = {
    "z": 0.02,
    "sol_ab": "as09",
    "imf": "kroupa",
    "imf_alpha": 2.35,
    "imf_m_low": 0.15,
    "imf_m_up": 100,
    "m_min": 0.98,
    "m_max": 40.0,
    "binary_fraction": constants.BIN_FRACTION,
    "dtd_sn": "rlp",
    "dtd_correction_factor": 1.0,
    "sn_yields": "iwa1998",
    "output_dir": "results",
    "matrix_headers": True,
    "return_fractions": False,
    "integration_step": "logt",
    "deprecation_warnings": True,
    "expelled_elements_filename": join(dirname(__file__), "sample_input", "expelled_elements"),
    "yield_corrections": {},
}

valid_values = {
    "imf": ["salpeter", "starburst", "chabrier", "ferrini", "kroupa", "miller_scalo", "maschberger"],
    "dtd_sn": ["rlp", "maoz", "castrillo", "greggio", "chen", "greggio-CDD04", "greggio-CDD1",
               "greggio-WDD04", "greggio-WDD1", "greggio-SDCH", "greggio-SDSCH",
               "strolger-fit1", "strolger-fit2", "strolger-fit3", "strolger-fit4", "strolger-optimized"],
    "sn_yields": ["iwa1998",
                  "sei2013",
                  "ln2020",
                  "ln2018-1", "ln2018-2", "ln2018-3",
                  "br2019-1", "br2019-2",
                  "gro2021-1", "gro2021-2",
                  "mor2018-1", "mor2018-2"],
    "sol_ab": ["ag89", "gs98", "as05", "as09", "he10", "lo19"],
    "integration_step": ["logt", "t", "two_steps_t", "fixed_n_steps"],
}

default_extraparams = {
    "integration_step": {
        "fixed_n_steps": {
            "integration_steps_stars_bigger_than_4Msun": 150,
            "integration_steps_stars_smaller_than_4Msun": 90
        },
        "logt": {
            "total_time_steps": 300,
        },
        "t": {
            "total_time_steps": 300,
        }
    }
}


def default_settings(params=default):
    default_params = default
    params = {**default_params, **params}

    for extraparam in default_extraparams.keys():
        if extraparam in params:
            if params[extraparam] in default_extraparams[extraparam].keys():
                default_params = {**default_params, **default_extraparams[extraparam][params[extraparam]]}

    return default_params


def validate(params):
    default_params = default_settings(params)
    params = {**default_params, **params}

    for param in valid_values.keys():
        if params[param] not in valid_values[param]:
            print(f"Provided value for {param} is incorrect.")
            print(f"  Valid values for {param} are: {valid_values[param]}")
            print(f"  Using default value: {default_params[param]}")
            params[param] = default_params[param]

    if params["m_max"] > max_mass_allowed(params["z"]):
        params["m_max"] = max_mass_allowed(params["z"])
        print(f"Maximum mass is bigger than the allowed mass for z: {params['z']}")
        print(f"  Using m_max value: {params['m_max']} solar masses")

    if params["imf"] == "starburst":
        params["imf_m_low"] = 1.0
        params["imf_m_up"] = 120.0

    if params["yield_corrections"] == {}:
        params.pop("yield_corrections")
    else:
        params["yield_corrections"] = validate_yield_corrections(params["yield_corrections"])

    invalid_params = params.keys() - default_params.keys()
    for invalid_param in invalid_params:
        print(f"Ignoring invalid setting: {invalid_param}")
        params.pop(invalid_param)

    print("")
    deprecation_warnings(params)

    return params


def validate_yield_corrections(corrections):
    valid_corrections = {}
    valid_elements = elements.Expelled.elements_list
    if type(corrections) is not dict:
        print("Yield corrections ignored")
        print("  Invalid format, it should be a dict of element: value pairs")
        print("  Valid values for element are: ")
        print(f"  {', '.join(valid_elements)}")
        print("")
        return valid_corrections

    normalized_input_corrections = dict((k.lower(), v) for k, v in corrections.items())

    for element in valid_elements:
        if element.lower() in normalized_input_corrections:
            value = normalized_input_corrections[element.lower()]
            if type(value) == int or type(value) == float:
                valid_corrections[element] = value
            else:
                print(f"Yield correction for {element} ignored: is not a number")

    invalid_corrections = normalized_input_corrections.keys() - [e.lower() for e in valid_elements]
    for invalid_correction in invalid_corrections:
        print(f"Yield correction for {invalid_correction} ignored: Invalid element")

    return valid_corrections


def deprecation_warnings(params):
    deprecation_warnings = []

    if params["deprecation_warnings"] is False:
        return []

    if params["deprecation_warnings"] == "test":
        deprecation_warnings.append("Deprecation warnings show here.")

    if deprecation_warnings:
        print("*** Deprecations Warning ***")
        for msg in deprecation_warnings:
            print("!! " + msg)
        print("")

    return deprecation_warnings



from bisect import bisect


class Expelled:

    elements_list = ["H", "D", "He3", "He4", "C12", "C13",
                     "N14p", "n.r.", "O16", "Ne", "Mg", "Si",
                     "S", "Ca", "Fe", "remnants", "C13s", "N14s"]
    mass_points = []
    by_mass = {}
    cri_lim_yields = False

    def __init__(self, expelled_elements_filename="expelled_elements"):
        self.mass_points = []
        self.by_mass = {}
        self.read_expelled_elements_file(expelled_elements_filename)

        upcased_filename = expelled_elements_filename.upper()
        if "CRI-LIM" in upcased_filename or "CRI_LIM" in upcased_filename:
            self.cri_lim_yields = True

    def read_expelled_elements_file(self, filename):
        """
        Reads a file of expelled elements per stellar mass.
        The file should include a row of data for each stellar mass.
        Structure of each row should be:
            - First column: stellar mass
            - 2nd to 19th columns: expelled mass of element i
                where i is in this ordered list:
                ["H", "D", "He3", "He4", "C12", "C13",
                 "N14primary", "n.r.", "O16", "Ne", "Mg", "Si",
                 "S", "Ca", "Fe", "remnants", "C13secondary", "N14secondary"]

        """

        expelled_data = open(filename, "r")

        for line in expelled_data:
            data_row = [max(0.0, float(data)) for data in line.split()]
            if data_row:
                mass = data_row.pop(0)  # the first column is the mass
                self.mass_points.append(mass)
                self.by_mass[mass] = dict(zip(self.elements_list, data_row))

        expelled_data.close()

    def for_mass(self, m, yield_corrections={}):
        """
        Interpolates expelled mass (per solar mass) for all elements for a given
        stellar mass, using the data from the class' expelled_elements input file.

        """

        index = bisect(self.mass_points, m)
        if index == len(self.mass_points):
            index -= 1

        mass_prev = self.mass_points[index - 1]
        mass_next = self.mass_points[index]
        elements_prev = self.by_mass[mass_prev]
        elements_next = self.by_mass[mass_next]
        interpolations = {"mass": m}
        p = (mass_next - m) / (mass_next - mass_prev)

        for element in self.elements_list:
            d = elements_next[element] - elements_prev[element]
            interpolations[element] = (elements_next[element] - (p * d)) / m
            if element in yield_corrections:
                interpolations[element] = interpolations[element] * yield_corrections[element]

        return interpolations



import math
import numpy as np
import starmatrix.constants as constants
import starmatrix.elements as elements
import starmatrix.matrix as matrix
from starmatrix.imfs import select_imf
from starmatrix.abundances import select_abundances
from starmatrix.dtds import select_dtd, dtd_correction
from starmatrix.functions import stellar_mass, stellar_lifetime, max_mass_allowed, return_fraction
from starmatrix.functions import total_energy_ejected, newton_cotes, global_imf, imf_supernovae_II


class Model:
    def __init__(self, settings={}):
        self.context = settings
        self.init_variables()

    def init_variables(self):
        self.initial_mass_function = select_imf(self.context["imf"], self.context)
        self.context["abundances"] = select_abundances(self.context["sol_ab"], float(self.context["z"]))
        self.context["expelled"] = elements.Expelled(expelled_elements_filename=self.context["expelled_elements_filename"])

        self.mass_intervals = []
        self.energies = []
        self.sn_Ia_rates = []

        self.z = self.context["z"]
        self.dtd = select_dtd(self.context["dtd_sn"])
        self.m_min = self.context["m_min"]
        self.m_max = self.context["m_max"]
        self.integration_step = self.context["integration_step"]
        self.total_time_steps = 0
        if "total_time_steps" in self.context:
            self.total_time_steps = self.context["total_time_steps"]

        self.bmaxm = constants.B_MAX / 2

    def run(self):
        self.explosive_nucleosynthesis()
        self.create_q_matrices()

    def create_q_matrices(self):
        q_sn_ia = matrix.q_sn(constants.CHANDRASEKHAR_LIMIT, feh=self.context["abundances"].feh(), sn_yields=self.context["sn_yields"])
        imf_sn_file = open(f"{self.context['output_dir']}/imf_supernova_rates", "w+")
        matrices_file = open(f"{self.context['output_dir']}/qm-matrices", "w+")
        if self.context["return_fractions"] is True:
            return_fraction_file = open(f"{self.context['output_dir']}/return_fractions", "w+")

        for i in range(0, self.total_time_steps):
            m_inf, m_sup = self.mass_intervals[i]
            q = np.zeros((constants.Q_MATRIX_ROWS, constants.Q_MATRIX_COLUMNS))
            phi, supernova_Ia_rates, supernova_II_rates, r = 0.0, 0.0, 0.0, 0.0

            if m_sup > constants.M_MIN and m_sup > m_inf:
                q += newton_cotes(
                    m_inf,
                    m_sup,
                    lambda m:
                        global_imf(m, self.initial_mass_function, self.context["binary_fraction"]) *
                        matrix.q(m, self.context)
                )

                supernova_Ia_rates = self.sn_Ia_rates[i] * self.initial_mass_function.stars_per_mass_unit * dtd_correction(self.context)
                q += q_sn_ia * supernova_Ia_rates

                phi = newton_cotes(
                    m_inf,
                    m_sup,
                    lambda m:
                        global_imf(m, self.initial_mass_function, self.context["binary_fraction"])
                )

                supernova_II_rates = newton_cotes(
                    m_inf,
                    m_sup,
                    lambda m:
                        imf_supernovae_II(m, self.initial_mass_function, self.context["binary_fraction"])
                )

                if self.context["return_fractions"] is True:
                    r = return_fraction(m_inf, m_sup, self.context["expelled"], self.initial_mass_function, self.context["binary_fraction"])

            np.savetxt(matrices_file, q, fmt="%15.10f", header=self._matrix_header(m_sup, m_inf))
            imf_sn_file.write(f"  {phi:.10f}  {supernova_Ia_rates:.10f}  {supernova_II_rates:.10f}  {self.energies[i]:.10f}\n")
            if self.context["return_fractions"] is True:
                return_fraction_file.write(f"{r:.10f}\n")

        matrices_file.close()
        imf_sn_file.close()
        if self.context["return_fractions"] is True:
            return_fraction_file.close()

    def explosive_nucleosynthesis(self):
        if self.integration_step == "logt":
            self.explosive_nucleosynthesis_step_logt()
        elif self.integration_step == "t":
            self.explosive_nucleosynthesis_step_t()
        elif self.integration_step == "two_steps_t":
            self.explosive_nucleosynthesis_two_steps_t()
        elif self.integration_step == "fixed_n_steps":
            steps_small_stars = self.context["integration_steps_stars_smaller_than_4Msun"]
            steps_massive_stars = self.context["integration_steps_stars_bigger_than_4Msun"]
            self.explosive_nucleosynthesis_fixed_n_steps(steps_massive_stars, steps_small_stars)
        else:
            raise ValueError("Invalid value for integration step. Should be one of: [logt, t, two_steps_t, fixed_n_steps]")

    def explosive_nucleosynthesis_step_logt(self):
        t_ini = stellar_lifetime(min(self.m_max, max_mass_allowed(self.z)), self.z)
        t_end = min(stellar_lifetime(self.m_min, self.z), constants.TOTAL_TIME)
        t_ini_log = math.log10(t_ini * 1e9)
        t_end_log = math.log10(t_end * 1e9)

        delta_t_log = (t_end_log - t_ini_log) / self.total_time_steps

        mass_intervals_file = open(f"{self.context['output_dir']}/mass_intervals", "w+")
        mass_intervals_file.write(" ".join([str(i) for i in [t_ini, t_end, self.total_time_steps, delta_t_log]]))

        for step in range(0, self.total_time_steps):
            t_inf_log = t_ini_log + (delta_t_log * step)
            t_sup_log = t_ini_log + (delta_t_log * (step + 1))

            t_inf = math.pow(10, t_inf_log - 9)
            t_sup = math.pow(10, t_sup_log - 9)

            m_inf = stellar_mass(t_sup, self.z)
            m_sup = stellar_mass(t_inf, self.z)

            mass_intervals_file.write('\n' + f'{m_sup:14.10f}  ' + f'{m_inf:14.10f}  ' + str(step + 1))

            self.mass_intervals.append([m_inf, m_sup])
            self.energies.append(total_energy_ejected(t_sup) - total_energy_ejected(t_inf))
            self.sn_Ia_rates.append(newton_cotes(t_inf, t_sup, self.dtd))

        mass_intervals_file.close()

    def explosive_nucleosynthesis_step_t(self):
        t_ini = stellar_lifetime(self.m_max, self.z)
        t_end = min(stellar_lifetime(self.m_min, self.z), constants.TOTAL_TIME)

        delta_t = (t_end - t_ini) / self.total_time_steps

        mass_intervals_file = open(f"{self.context['output_dir']}/mass_intervals", "w+")
        mass_intervals_file.write(" ".join([str(i) for i in [t_ini, t_end, self.total_time_steps, delta_t]]))

        for step in range(0, self.total_time_steps):
            t_inf = t_ini + (delta_t * step)
            t_sup = t_ini + (delta_t * (step + 1))

            m_inf = stellar_mass(t_sup, self.z)
            m_sup = stellar_mass(t_inf, self.z)

            mass_intervals_file.write('\n' + f'{m_sup:14.10f}  ' + f'{m_inf:14.10f}  ' + str(step + 1))

            self.mass_intervals.append([m_inf, m_sup])
            self.energies.append(total_energy_ejected(t_sup) - total_energy_ejected(t_inf))
            self.sn_Ia_rates.append(newton_cotes(t_inf, t_sup, self.dtd))

        mass_intervals_file.close()

    def explosive_nucleosynthesis_two_steps_t(self):
        t_ini = stellar_lifetime(self.m_max, self.z)
        t_limit_massive = stellar_lifetime(4.0, self.z)
        t_end = min(stellar_lifetime(self.m_min, self.z), constants.TOTAL_TIME)

        delta_t_1 = stellar_lifetime(100.0, self.z) / 2
        delta_t_2 = 50 * delta_t_1

        steps_with_delta_t_1 = math.ceil((t_limit_massive - t_ini) / delta_t_1)
        t_ini_for_delta_2 = t_ini + (delta_t_1 * steps_with_delta_t_1)
        steps_with_delta_t_2 = math.ceil((t_end - t_ini_for_delta_2) / delta_t_2)
        delta_t_2 = (t_end - t_ini_for_delta_2) / steps_with_delta_t_2
        self.total_time_steps = steps_with_delta_t_1 + steps_with_delta_t_2

        mass_intervals_file = open(f"{self.context['output_dir']}/mass_intervals", "w+")
        mass_intervals_file.write(" ".join([str(i) for i in [t_ini, t_end, steps_with_delta_t_1, steps_with_delta_t_2, delta_t_1, delta_t_2]]))

        for step in range(0, steps_with_delta_t_1):
            t_inf = t_ini + (delta_t_1 * step)
            t_sup = t_ini + (delta_t_1 * (step + 1))

            m_inf = stellar_mass(t_sup, self.z)
            m_sup = stellar_mass(t_inf, self.z)

            mass_intervals_file.write('\n' + f'{m_sup:14.10f}  ' + f'{m_inf:14.10f}  ' + str(step + 1))

            self.mass_intervals.append([m_inf, m_sup])
            self.energies.append(total_energy_ejected(t_sup) - total_energy_ejected(t_inf))
            self.sn_Ia_rates.append(newton_cotes(t_inf, t_sup, self.dtd))

        for step in range(0, steps_with_delta_t_2):
            t_inf = t_ini_for_delta_2 + (delta_t_2 * step)
            t_sup = t_ini_for_delta_2 + (delta_t_2 * (step + 1))

            m_inf = stellar_mass(t_sup, self.z)
            m_sup = stellar_mass(t_inf, self.z)

            mass_intervals_file.write('\n' + f'{m_sup:14.10f}  ' + f'{m_inf:14.10f}  ' + str(step + 1))

            self.mass_intervals.append([m_inf, m_sup])
            self.energies.append(total_energy_ejected(t_sup) - total_energy_ejected(t_inf))
            self.sn_Ia_rates.append(newton_cotes(t_inf, t_sup, self.dtd))

        mass_intervals_file.close()

    def explosive_nucleosynthesis_fixed_n_steps(self, n_massive, n_small):
        t_ini = stellar_lifetime(self.m_max, self.z)
        t_limit_massive = stellar_lifetime(4.0, self.z)
        t_end = min(stellar_lifetime(self.m_min, self.z), constants.TOTAL_TIME)

        delta_t_1 = (t_limit_massive - t_ini) / n_massive
        delta_t_2 = (t_end - t_limit_massive) / n_small

        self.total_time_steps = n_massive + n_small

        mass_intervals_file = open(f"{self.context['output_dir']}/mass_intervals", "w+")
        mass_intervals_file.write(" ".join([str(i) for i in [t_ini, t_end, n_massive, n_small, delta_t_1, delta_t_2]]))

        for step in range(0, n_massive):
            t_inf = t_ini + (delta_t_1 * step)
            t_sup = t_ini + (delta_t_1 * (step + 1))

            m_inf = stellar_mass(t_sup, self.z)
            m_sup = stellar_mass(t_inf, self.z)

            mass_intervals_file.write('\n' + f'{m_sup:14.10f}  ' + f'{m_inf:14.10f}  ' + str(step + 1))

            self.mass_intervals.append([m_inf, m_sup])
            self.energies.append(total_energy_ejected(t_sup) - total_energy_ejected(t_inf))
            self.sn_Ia_rates.append(newton_cotes(t_inf, t_sup, self.dtd))

        for step in range(0, n_small):
            t_inf = t_limit_massive + (delta_t_2 * step)
            t_sup = t_limit_massive + (delta_t_2 * (step + 1))

            m_inf = stellar_mass(t_sup, self.z)
            m_sup = stellar_mass(t_inf, self.z)

            mass_intervals_file.write('\n' + f'{m_sup:14.10f}  ' + f'{m_inf:14.10f}  ' + str(step + 1))

            self.mass_intervals.append([m_inf, m_sup])
            self.energies.append(total_energy_ejected(t_sup) - total_energy_ejected(t_inf))
            self.sn_Ia_rates.append(newton_cotes(t_inf, t_sup, self.dtd))

        mass_intervals_file.close()

    def _matrix_header(self, m_sup, m_inf):
        if self.context["matrix_headers"] is True:
            return f"Q matrix for mass interval: [{m_sup}, {m_inf}]"
        else:
            return ""



import pytest
import starmatrix.settings as settings
from starmatrix.functions import max_mass_allowed


def test_defaults():
    assert settings.default is not None
    for setting in ["z", "sol_ab", "imf", "imf_alpha",
                    "m_max", "binary_fraction", "dtd_sn", "sn_yields",
                    "output_dir", "expelled_elements_filename"]:
        assert settings.default[setting] is not None


def test_default_settings_with_no_params():
    assert settings.default_settings() == settings.default_settings(settings.default)


def test_default_settings_adds_extra_params():
    test_settings = {"integration_step": "fixed_n_steps"}
    expected_values = settings.default_extraparams["integration_step"]["fixed_n_steps"]
    complete_default_settings = settings.default_settings(test_settings)

    for k in ["integration_steps_stars_bigger_than_4Msun", "integration_steps_stars_smaller_than_4Msun"]:
        assert (k in settings.default) is False
        assert (k in complete_default_settings) is True
        assert complete_default_settings[k] == expected_values[k]


def test_default_settings_adds_extra_params_only_if_needed():
    settings.default_extraparams["inexistent_setting"] = {"a": 34}
    params = settings.default_settings({"integration_step": "testing"})
    assert(params) == settings.default
    assert("inexistent_setting" in params) is False


def test_validate_with_valid_values():
    valid_values = {
        "z": 0.033,
        "imf": "miller_scalo",
        "imf_m_low": 0.3,
        "dtd_sn": "castrillo",
        "sol_ab": "ag89",
        "sn_yields": "iwa1998",
        "m_max": 40.0,
        "output_dir": "testing"
    }

    params = settings.validate(valid_values)

    for param in ["z", "imf", "imf_m_low", "dtd_sn", "sol_ab", "m_max", "output_dir"]:
        assert params[param] == valid_values[param]


def test_validate_with_invalid_values():
    invalid_values = {
        "imf": "wrong name",
        "dtd_sn": "whatever",
        "sol_ab": "invalid",
        "sn_yields": "incorrect",
        "m_max": 3.3
    }

    params = settings.validate(invalid_values)

    for param in ["imf", "dtd_sn", "sol_ab", "sn_yields"]:
        assert params[param] != invalid_values[param]
        assert params[param] == settings.default[param]


def test_validate_max_mass_with_valid_values():
    params = settings.validate({"m_max": 33})
    assert params["m_max"] == 33


def test_validate_max_mass_with_invalid_values():
    for z in [0.001, 0.008, 0.2, 0.33, 0.5]:
        params = settings.validate({"m_max": 300, "z": z})
        assert params["m_max"] == max_mass_allowed(z)


def test_validate_mass_limits_for_starburst_imf():
    for imf in settings.valid_values["imf"]:
        params = settings.validate({"imf": imf, "imf_m_low": 7, "imf_m_up": 40})
        if imf == "starburst":
            assert params["imf_m_low"] == 1
            assert params["imf_m_up"] == 120
        else:
            assert params["imf_m_low"] == 7
            assert params["imf_m_up"] == 40


def test_validate_with_invalid_settings():
    params = settings.validate({"imf_m_low": 7, "invalid_setting": 47})

    assert(params["imf_m_low"]) == 7
    assert("invalid_setting" in params) is False


def test_validate_with_yield_corrections():
    params = settings.validate({"yield_corrections": {"Mg": 3, "Aluminium": 4}})
    assert(params["yield_corrections"]) == {"Mg": 3}


def test_validate_yield_corrections_with_invalid_input():
    no_dict_input = settings.validate_yield_corrections([])
    invalid_value_input = settings.validate_yield_corrections({"H": "A lot"})

    assert(no_dict_input) == {}
    assert(invalid_value_input) == {}


def test_validate_yield_corrections():
    corrections = {"H": 1.3, "mg": 2.1, "fE": 4, "AllTheRest": 7.8, "Si": "Wrong"}
    valid_corrections = settings.validate_yield_corrections(corrections)

    assert(valid_corrections) == {"H": 1.3, "Mg": 2.1, "Fe": 4}


def test_deprecation_warnings():
    dw = settings.deprecation_warnings({"deprecation_warnings": False})
    assert len(dw) == 0

    dw = settings.deprecation_warnings({"deprecation_warnings": "test"})
    assert len(dw) == 1



import numpy as np
import starmatrix.constants as constants
import starmatrix.functions as functions
import starmatrix.supernovae as sn


def empty_q_matrix():
    return np.zeros((15, 15))


def q_index(element):
    q_elements = ["H", "D", "He3", "He4", "C12", "O16", "N14", "C13", "nr", "Ne", "Mg", "Si", "S", "Ca", "Fe"]
    return q_elements.index(element)


def q(m, settings={}):
    """
    Compute the Q Matrix of elements for a given mass (without supernovae)

    The element production matrix has this structure:
        H  D  He3  He4  C12  O16  N14  C13  nr  Ne  Mg  Si  S  Ca  Fe
    H
    D
    He3
    He4
    C12
    O16
    N14
    C13
    nr
    Ne
    Mg
    Si
    S
    Ca
    Fe

    So element Q(1,1) (internally q(0,0) as numpy index starts at 0) is the H produced from H,
    and Q(14,4) is the Calcium created from Helium 4.
    Returned matrix is cropped to [constants.Q_MATRIX_ROWS, constants.Q_MATRIX_COLUMNS]

    """

    q = empty_q_matrix()
    if m < constants.M_MIN:
        return resize_matrix(q)

    z = settings["z"]
    expelled = settings["expelled"]

    # Apply corrections if present
    yield_corrections = {}
    if "yield_corrections" in settings:
        yield_corrections = settings["yield_corrections"]

    elements = expelled.for_mass(m, yield_corrections)
    abundances = settings["abundances"].abundance()

    # CRI-LIM correction
    if expelled.cri_lim_yields:
        abundances = settings["abundances"].corrected_abundance_CRI_LIM()

    h_he = abundances["H"] + abundances["He4"]
    remnant = elements["remnants"]
    he_core = 1 - (elements["H"] / abundances["H"])
    co_core = ((he_core * abundances["H"]) + abundances["He4"] - elements["He4"]) / h_he
    new_metals_ejected = max(co_core - remnant, 0.0)

    fractional_abundances = {
        "D":   0.0,
        "He3": 0.0,
        "N":   0.0,
        "C":   0.0,
        "C13": 0.0,
        "O":   0.0,
        "Ne":  0.0,
        "Mg":  0.0,
        "Si":  0.0,
        "S":   0.0,
        "Ca":  0.0,
        "Fe":  0.0
    }

    # Secondary production of N and C
    secondary_n_core = (
        (elements["N14s"] / (abundances["C"] + abundances["C13"] + abundances["O"])) -
        ((1 - co_core) * abundances["N"] / (abundances["C"] + abundances["C13"] + abundances["O"])) +
        co_core
    )
    secondary_c13_core = (
        (elements["C13s"] / abundances["C"]) -
        ((1 - secondary_n_core) * (abundances["C13"] / abundances["C"])) +
        secondary_n_core
    )

    if new_metals_ejected > 0:
        fractional_abundances["N"] = elements["N14p"] / (h_he * new_metals_ejected)
        fractional_abundances["C13"] = elements["C13"] / (h_he * new_metals_ejected)
        fractional_abundances["C"] = (
            (elements["C12"] / (h_he * new_metals_ejected)) -
            ((1 - secondary_c13_core) * abundances["C"] / (h_he * new_metals_ejected))
        )
        fractional_abundances["O"] = (
            (elements["O16"] / (h_he * new_metals_ejected)) -
            ((1 - secondary_n_core) * abundances["O"] / (h_he * new_metals_ejected))
        )

        for element in ["Ne", "Mg", "Si", "S", "Ca", "Fe"]:
            fractional_abundances[element] = (
                (elements[element] - ((1 - remnant) * abundances[element])) /
                (h_he * new_metals_ejected)
            )

    # Make sure all values are in [0, 1] and normalize:
    for key, value in fractional_abundances.items():
        fractional_abundances[key] = functions.value_in_interval(value, [0.0, 1.0])

    total_abundances = sum(fractional_abundances.values())
    if total_abundances > 1:
        for key, value in fractional_abundances.items():
            fractional_abundances[key] = value / total_abundances

    # He3 core:
    if m <= 3:
        he3_core = he_core
    elif 3 < m <= 8:
        he3_core = 0.282 + 0.026 * m
    elif 8 < m <= 15:
        he3_core = 0.33 + 0.02 * m
    elif 15 < m <= 25:
        he3_core = 0.525 + 0.007 * m
    elif 25 < m <= 50:
        he3_core = 0.63 + 0.00288 * m
    elif 50 < m:
        he3_core = 0.73 + 0.0008 * m

    # Omega He3:
    if constants.M_MIN <= m < 2:
        w3 = (-3.47e-4 * m) + 7.79e-4
    elif 2 <= m <= 3:
        w3 = (-4.43e-5 * m) + 1.74e-4
    elif 3 <= m <= 5:
        w3 = (-1.15e-5 * m) + 7.53e-5
    else:
        w3 = 0.

    fractional_abundances["He3"] = w3 * (1 - remnant) / abundances["H"]

    # Q(i,j) values:
    q[0, 0] = 1 - he_core - fractional_abundances["He3"]
    q[0, 1] = -0.5 * (1 - remnant)
    q[2, 0] = fractional_abundances["He3"]
    q[2, 1] = 1.5 * (1 - he3_core)
    q[2, 2] = 1 - he3_core
    q[3, 0] = he_core - co_core
    q[3, 1] = 1.5 * (he3_core - co_core)
    q[3, 2] = he3_core - co_core
    q[3, 3] = 1 - co_core

    q[4, 0] = fractional_abundances["C"] * new_metals_ejected
    q[5, 0] = fractional_abundances["O"] * new_metals_ejected
    q[6, 0] = fractional_abundances["N"] * new_metals_ejected
    q[7, 0] = fractional_abundances["C13"] * new_metals_ejected

    q[9, 0] = fractional_abundances["Ne"] * new_metals_ejected
    q[10, 0] = fractional_abundances["Mg"] * new_metals_ejected
    q[11, 0] = fractional_abundances["Si"] * new_metals_ejected
    q[12, 0] = fractional_abundances["S"] * new_metals_ejected
    q[13, 0] = fractional_abundances["Ca"] * new_metals_ejected
    q[14, 0] = fractional_abundances["Fe"] * new_metals_ejected

    # C12  O16  N14  C13  Ne  Mg  Si  S  Ca  Fe
    for i in [4, 5, 6, 7, 9, 10, 11, 12, 13, 14]:
        q[i, 1] = 1.5 * q[i, 0]
        q[i, 2] = q[i, 0]
        q[i, 3] = q[i, 0]

    # n.r.:
    q[8][4] = new_metals_ejected
    q[8][5] = new_metals_ejected
    q[8][6] = new_metals_ejected
    q[8][7] = new_metals_ejected

    # Diagonal
    for i in range(8, 15):
        q[i][i] = 1 - remnant

    # C-N-O cycle:
    q[4, 4] = 1 - secondary_c13_core
    q[5, 5] = 1 - secondary_n_core
    q[6, 6] = 1 - co_core
    q[7, 7] = 1 - secondary_n_core
    q[6, 4] = secondary_n_core - co_core
    q[6, 5] = secondary_n_core - co_core
    q[6, 7] = secondary_n_core - co_core
    q[7, 4] = secondary_c13_core - secondary_n_core

    # No negative values allowed except for H-D (q(0,1)):
    for i in range(0, 15):
        for j in range(0, 15):
            if q[i, j] <= 0.0 and (i != 0 and j != 1):
                q[i, j] = 0.0

    return resize_matrix(q)


def q_sn(m, feh=0.0, sn_yields="iwa1998"):
    """
    Compute the Q Matrix of elements coming from Supernova events
    Supernovae type can be specified as one of: [sn_ia, sn_ib]

    Returned matrix is cropped to [constants.Q_MATRIX_ROWS, constants.Q_MATRIX_COLUMNS]

    """

    q = empty_q_matrix()
    if m < constants.M_MIN:
        return resize_matrix(q)

    ejected = sn.empty_yields_set()
    sn_ejections = sn.yields(sn_yields, feh)
    den = 0.99 * m

    for element in ejected.keys():
        ejected[element] = sn_ejections[element] / den
        for i in range(0, 4):
            q[q_index(element), i] = ejected[element]

    # Remnants = Stellar mass - Ejected mass
    remnant = (m - sum(sn_ejections.values())) / m
    remnant = functions.value_in_interval(remnant, [0.0, 1.0])

    # Diagonal for [nr  Ne  Mg  Si  S  Ca  Fe]
    for i in range(8, 15):
        q[i, i] = 1.0 - remnant

    return resize_matrix(q)


def resize_matrix(complete_matrix):
    return complete_matrix[0:constants.Q_MATRIX_ROWS, 0:constants.Q_MATRIX_COLUMNS]



# Q-matrices size:
Q_MATRIX_ROWS = 15
Q_MATRIX_COLUMNS = 9

# Mass range config:
M_MIN = 0.8         # Minimum mass
CHANDRASEKHAR_LIMIT = 1.4

# Binaries params:
M_SNII = 8.0         # Lower limit for binaries in Supernovae II
B_MIN = 3.0          # Lower limit for binaries in Supernovae I
B_MAX = 16.0         # Upper limit for binaries in Supernovae I
BIN_FRACTION = 0.15  # Fraction of binaries

# Model calculations params:
TOTAL_TIME = 13.25   # Total integration time in Gigayears
